import re
from typing import List, Tuple, Optional, cast
import sys

import numpy as np
import numpy.typing as npt
from PIL import Image, UnidentifiedImageError
from rclip.utils import helpers
from importlib.metadata import version
import open_clip

QUERY_WITH_MULTIPLIER_RE = re.compile(r"^(?P<multiplier>(\d+(\.\d+)?|\.\d+|\d+\.)):(?P<query>.+)$")
QueryWithMultiplier = Tuple[float, str]
FeatureVector = npt.NDArray[np.float32]

TEXT_ONLY_SUPPORTED_MODELS = [
  {
    "model_name": "ViT-B-32-quickgelu",
    "checkpoint_name": "openai",
  }
]


def get_open_clip_version():
  return version("open_clip_torch")


class Model:
  VECTOR_SIZE = 512
  _model_name = "ViT-B-32-quickgelu"
  _checkpoint_name = "openai"

  def __init__(self, device: str = "cpu"):
    self._device = device

    self._model_var = None
    self._model_text_var = None
    self._preprocess_var = None
    self._tokenizer_var = None

    self._text_model_path = helpers.get_app_datadir() / f"{self._model_name}_{self._checkpoint_name}_text.pth"
    self._text_model_version_path = (
      helpers.get_app_datadir() / f"{self._model_name}_{self._checkpoint_name}_text.version"
    )

  @property
  def _tokenizer(self):
    if not self._tokenizer_var:
      self._tokenizer_var = open_clip.get_tokenizer(self._model_name)
    return self._tokenizer_var

  def _load_model(self):
    self._model_var, _, self._preprocess_var = open_clip.create_model_and_transforms(
      self._model_name,
      pretrained=self._checkpoint_name,
      device=self._device,
    )
    self._model_text_var = None

    if {
      "model_name": self._model_name,
      "checkpoint_name": self._checkpoint_name,
    } in TEXT_ONLY_SUPPORTED_MODELS and self._should_update_text_model():
      import torch

      model_text = self._get_text_model(cast(open_clip.CLIP, self._model_var))
      torch.save(model_text, self._text_model_path)

      with self._text_model_version_path.open("w") as f:
        f.write(get_open_clip_version())

  @staticmethod
  def _get_text_model(model: open_clip.CLIP):
    import copy

    model_text = copy.deepcopy(model)
    model_text.visual = None  # type: ignore
    return model_text

  def _should_update_text_model(self):
    if not self._text_model_path.exists():
      return True

    if not self._text_model_version_path.exists():
      return True

    with self._text_model_version_path.open("r") as f:
      text_model_version = f.read().strip()

    # to be safe, update the text model on open_clip update (which could update the base model)
    return get_open_clip_version() != text_model_version

  @property
  def _model(self):
    if not self._model_var:
      self._load_model()
    return cast(open_clip.CLIP, self._model_var)

  @property
  def _model_text(self):
    if self._model_var:
      return self._model_var

    if self._model_text_var:
      return self._model_text_var

    if self._text_model_path.exists() and not self._should_update_text_model():
      import torch

      self._model_text_var = torch.load(self._text_model_path, weights_only=False)
      return self._model_text_var

    if not self._model_var:
      self._load_model()

    return cast(open_clip.CLIP, self._model_var)

  @property
  def _preprocess(self):
    from torchvision.transforms import Compose

    if not self._preprocess_var:
      self._load_model()
    return cast(Compose, self._preprocess_var)

  def compute_image_features(self, images: List[Image.Image]) -> npt.NDArray[np.float32]:
    import torch

    images_preprocessed = torch.stack(cast(list[torch.Tensor], [self._preprocess(thumb) for thumb in images])).to(
      self._device
    )
    with torch.no_grad():
      image_features = self._model.encode_image(images_preprocessed)
      image_features /= image_features.norm(dim=-1, keepdim=True)
    return image_features.cpu().numpy()

  def compute_text_features(self, text: List[str]) -> npt.NDArray[np.float32]:
    import torch

    with torch.no_grad():
      text_features = self._model_text.encode_text(self._tokenizer(text).to(self._device))
      text_features /= text_features.norm(dim=-1, keepdim=True)
    return text_features.cpu().numpy()

  @staticmethod
  def _extract_query_multiplier(query: str) -> QueryWithMultiplier:
    match = QUERY_WITH_MULTIPLIER_RE.match(query)
    if not match:
      return 1.0, query
    multiplier = float(match.group("multiplier"))
    query = match.group("query")
    return multiplier, query

  @staticmethod
  def _group_queries_by_type(
    queries: List[str],
  ) -> Tuple[List[QueryWithMultiplier], List[QueryWithMultiplier], List[QueryWithMultiplier]]:
    phrase_queries: List[Tuple[float, str]] = []
    local_file_queries: List[Tuple[float, str]] = []
    url_queries: List[Tuple[float, str]] = []
    for query in queries:
      multiplier, query = Model._extract_query_multiplier(query)
      if helpers.is_http_url(query):
        url_queries.append((multiplier, query))
      elif helpers.is_file_path(query):
        local_file_queries.append((multiplier, query))
      else:
        phrase_queries.append((multiplier, query))
    return phrase_queries, local_file_queries, url_queries

  def compute_features_for_queries(self, queries: List[str]) -> FeatureVector:
    text_features: Optional[FeatureVector] = None
    image_features: Optional[FeatureVector] = None
    phrases, files, urls = self._group_queries_by_type(queries)

    # process images first to avoid loading BOTH full and text-only models
    # if we need to process images, we will load the full model, and the text processing logic will use it, too
    # if we don't need to process images, we will skip loading the full model, and the text processing
    # logic will load the text-only model

    if files or urls:
      file_multipliers, file_paths = cast(Tuple[Tuple[float], Tuple[str]], zip(*(files))) if files else ((), ())
      url_multipliers, url_paths = cast(Tuple[Tuple[float], Tuple[str]], zip(*(urls))) if urls else ((), ())
      try:
        images = [helpers.download_image(q) for q in url_paths] + [helpers.read_image(q) for q in file_paths]
      except FileNotFoundError as e:
        print(f'File "{e.filename}" not found. Check if you have typos in the filename.')
        sys.exit(1)
      except UnidentifiedImageError as e:
        print(f'File "{e.filename}" is not an image. You can only use image files or text as queries.')
        sys.exit(1)
      image_multipliers = np.array(url_multipliers + file_multipliers)
      image_features = np.add.reduce(self.compute_image_features(images) * image_multipliers.reshape(-1, 1))

    if phrases:
      phrase_multipliers, phrase_queries = cast(Tuple[Tuple[float], Tuple[str]], zip(*phrases))
      phrase_multipliers_np = np.array(phrase_multipliers).reshape(-1, 1)
      text_features = np.add.reduce(self.compute_text_features([*phrase_queries]) * phrase_multipliers_np)

    if text_features is not None and image_features is not None:
      return text_features + image_features
    elif text_features is not None:
      return text_features
    elif image_features is not None:
      return image_features
    else:
      return np.zeros(Model.VECTOR_SIZE, dtype=np.float32)

  def compute_similarities_to_text(
    self, item_features: FeatureVector, positive_queries: List[str], negative_queries: List[str]
  ) -> List[Tuple[float, int]]:
    positive_features = self.compute_features_for_queries(positive_queries)
    negative_features = self.compute_features_for_queries(negative_queries)

    features = positive_features - negative_features

    similarities = features @ item_features.T
    sorted_similarities = sorted(zip(similarities, range(item_features.shape[0])), key=lambda x: x[0], reverse=True)

    return sorted_similarities



import os
import sys


def is_snap():
  return bool(os.getenv("SNAP"))


def check_snap_permissions(directory: str):
  try:
    any(os.scandir(directory))
  except PermissionError:
    homedir = os.getenv("SNAP_REAL_HOME")
    if not homedir:
      print(
        "SNAP_REAL_HOME environment variable is not set."
        " Please, report the issue to the rclip project on"
        " GitHub https://github.com/yurijmikhalevich/rclip/issues."
      )
      sys.exit(1)
    if directory == homedir or directory.startswith(homedir + os.sep):
      print(
        "rclip doesn't have access to the current directory."
        " You can resolve this issue by running:"
        "\n\n\tsudo snap connect rclip:home\n\n"
        "This command will grant rclip the necessary access to the home directory."
        " Afterward, you can try again."
      )
      sys.exit(1)
    if directory == "/media" or directory.startswith("/media" + os.sep):
      print(
        "rclip doesn't have access to the current directory."
        " You can resolve this issue by running:"
        "\n\n\tsudo snap connect rclip:removable-media\n\n"
        'This command will grant rclip the necessary access to the "/media" directory.'
        " Afterward, you can try again."
      )
      sys.exit(1)
    print(
      'Running rclip outside of the home or "/media" directories is not supported by its snap version.'
      ' If you want to use rclip outside of home or "/media",'
      " file an issue in the rclip project on GitHub https://github.com/yurijmikhalevich/rclip/issues,"
      " describe your use case, and consider alternative rclip installation"
      " options https://github.com/yurijmikhalevich/rclip#linux."
    )
    sys.exit(1)



import sys


IS_MACOS = sys.platform == "darwin"
IS_LINUX = sys.platform.startswith("linux")
IS_WINDOWS = sys.platform == "win32" or sys.platform == "cygwin"

# these images are always processed
IMAGE_EXT = ["jpg", "jpeg", "png", "webp"]
# RAW images are processed only if there is no processed image alongside it
IMAGE_RAW_EXT = ["arw", "cr2"]



import argparse
import os
import pathlib
import textwrap
from typing import IO, cast
from PIL import Image, UnidentifiedImageError
import re
import numpy as np
import rawpy
import requests
import sys
from importlib.metadata import version

from rclip.const import IMAGE_RAW_EXT, IS_LINUX, IS_MACOS, IS_WINDOWS


MAX_DOWNLOAD_SIZE_BYTES = 50_000_000
DOWNLOAD_TIMEOUT_SECONDS = 60
WIN_ABSOLUTE_FILE_PATH_REGEX = re.compile(r"^[a-z]:\\", re.I)
DEFAULT_TERMINAL_TEXT_WIDTH = 100


def __get_system_datadir() -> pathlib.Path:
  """
  Returns a parent directory path
  where persistent application data can be stored.

  - linux: ~/.local/share
  - macOS: ~/Library/Application Support
  - windows: C:/Users/<USER>/AppData/Roaming
  """

  home = pathlib.Path.home()

  if IS_WINDOWS:
    return home / "AppData/Roaming"
  elif IS_LINUX:
    return home / ".local/share"
  elif IS_MACOS:
    return home / "Library/Application Support"

  raise NotImplementedError(f'"{sys.platform}" is not supported')


def get_app_datadir() -> pathlib.Path:
  app_datadir = os.getenv("RCLIP_DATADIR")
  if app_datadir:
    app_datadir = pathlib.Path(app_datadir)
  else:
    app_datadir = __get_system_datadir() / "rclip"
  os.makedirs(app_datadir, exist_ok=True)
  return app_datadir


def positive_int_arg_type(arg: str) -> int:
  arg_int = int(arg)
  if arg_int < 1:
    raise argparse.ArgumentTypeError("should be >0")
  return arg_int


def get_terminal_text_width() -> int:
  try:
    computed_width = min(DEFAULT_TERMINAL_TEXT_WIDTH, os.get_terminal_size().columns - 2)
    if computed_width < 20:
      return DEFAULT_TERMINAL_TEXT_WIDTH
    return computed_width
  except OSError:
    return DEFAULT_TERMINAL_TEXT_WIDTH


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
  def __init__(self, prog: str, indent_increment: int = 2, max_help_position: int = 24) -> None:
    text_width = get_terminal_text_width()
    super().__init__(prog, indent_increment, max_help_position, width=text_width)


def init_arg_parser() -> argparse.ArgumentParser:
  text_width = get_terminal_text_width()
  parser = argparse.ArgumentParser(
    formatter_class=HelpFormatter,
    prefix_chars="-+",
    description="rclip is an AI-powered command-line photo search tool",
    epilog="hints:\n"
    + textwrap.fill(
      '- relative file path should be prefixed with ./, e.g. "./cat.jpg", not "cat.jpg"',
      initial_indent="  ",
      subsequent_indent="    ",
      width=text_width,
    )
    + "\n"
    + textwrap.fill(
      '- any query can be prefixed with a multiplier, e.g. "2:cat", "0.5:./cat-sleeps-on-a-chair.jpg";'
      " adding a multiplier is especially useful when combining image and text queries because"
      " image queries are usually weighted more than text ones",
      initial_indent="  ",
      subsequent_indent="    ",
      width=text_width,
    )
    + "\n\n"
    "get help:\n"
    "  https://github.com/yurijmikhalevich/rclip/discussions/new/choose\n\n",
  )
  version_str = f"rclip {version('rclip')}"
  parser.add_argument("--version", "-v", action="version", version=version_str, help=f'prints "{version_str}"')
  parser.add_argument("query", help="a text query or a path/URL to an image file")
  parser.add_argument(
    "--add",
    "-a",
    "+",
    metavar="QUERY",
    action="append",
    default=[],
    help='a text query or a path/URL to an image file to add to the "original" query, can be used multiple times',
  )
  parser.add_argument(
    "--subtract",
    "--sub",
    "-s",
    "-",
    metavar="QUERY",
    action="append",
    default=[],
    help='a text query or a path/URL to an image file to subtract from the "original" query,'
    " can be used multiple times",
  )
  parser.add_argument(
    "--top", "-t", type=positive_int_arg_type, default=10, help="number of top results to display; default: 10"
  )
  display_mode_group = parser.add_mutually_exclusive_group()
  display_mode_group.add_argument(
    "--preview",
    "-p",
    action="store_true",
    default=False,
    help="preview results in the terminal (supported in iTerm2, Konsole 22.04+, wezterm, Mintty, mlterm)",
  )
  display_mode_group.add_argument(
    "--filepath-only",
    "-f",
    action="store_true",
    default=False,
    help="outputs only filepaths",
  )
  parser.add_argument(
    "--preview-height",
    "-H",
    metavar="PREVIEW_HEIGHT_PX",
    action="store",
    type=int,
    default=400,
    help="preview height in pixels; default: 400",
  )
  parser.add_argument(
    "--no-indexing",
    "--skip-index",
    "--skip-indexing",
    "-n",
    action="store_true",
    default=False,
    help="allows to skip updating the index if no images were added, changed, or removed",
  )
  parser.add_argument(
    "--indexing-batch-size",
    "-b",
    type=positive_int_arg_type,
    default=8,
    help="the size of the image batch used when updating the search index;"
    " larger values may improve the indexing speed a bit on some hardware but will increase RAM usage; default: 8",
  )
  parser.add_argument(
    "--exclude-dir",
    action="append",
    help="dir to exclude from search, can be used multiple times;"
    ' adding this argument overrides the default of ("@eaDir", "node_modules", ".git");'
    " WARNING: the default will be removed in v2",
  )
  parser.add_argument(
    "--experimental-raw-support",
    action="store_true",
    default=False,
    help="enables support for RAW images (only ARW and CR2 are supported)",
  )
  if IS_MACOS:
    if is_mps_available():
      parser.add_argument(
        "--device", "-d", default="mps", choices=["cpu", "mps"], help="device to run on; default: mps"
      )
  return parser


def is_mps_available() -> bool:
  if not IS_MACOS:
    return False
  import torch.backends.mps

  if not torch.backends.mps.is_available():
    return False
  try:
    import torch

    # on some systems, specifically in GHA
    # torch.backends.mps.is_available() returns True, but using the mps backend fails
    torch.ones(1, device="mps")
    return True
  except RuntimeError:
    return False


# See: https://meta.wikimedia.org/wiki/User-Agent_policy
def download_image(url: str) -> Image.Image:
  headers = {"User-agent": "rclip - (https://github.com/yurijmikhalevich/rclip)"}
  check_size = requests.request("HEAD", url, headers=headers, timeout=60)
  if length := check_size.headers.get("Content-Length"):
    if int(length) > MAX_DOWNLOAD_SIZE_BYTES:
      raise ValueError(f"Avoiding download of large ({length} byte) file.")
  img = Image.open(
    cast(IO[bytes], requests.get(url, headers=headers, stream=True, timeout=DOWNLOAD_TIMEOUT_SECONDS).raw)
  )
  return img


def get_file_extension(path: str) -> str:
  return os.path.splitext(path)[1].lower()[1:]


def read_raw_image_file(path: str):
  raw = rawpy.imread(path)
  rgb = raw.postprocess()
  return Image.fromarray(np.array(rgb))


def read_image(query: str) -> Image.Image:
  path = str.removeprefix(query, "file://")
  try:
    file_ext = get_file_extension(path)
    if file_ext in IMAGE_RAW_EXT:
      image = read_raw_image_file(path)
    else:
      image = Image.open(path)
  except UnidentifiedImageError as e:
    # by default the filename on the UnidentifiedImageError is None
    e.filename = path
    raise e
  return image


def is_http_url(path: str) -> bool:
  return path.startswith("https://") or path.startswith("http://")


def is_file_path(path: str) -> bool:
  return (
    path.startswith("/")
    or path.startswith("file://")
    or path.startswith("./")
    or WIN_ABSOLUTE_FILE_PATH_REGEX.match(path) is not None
  )



import base64
from io import BytesIO
import os
from PIL import Image

from rclip.utils.helpers import read_image


def _get_start_sequence():
  term_env_var = os.getenv("TERM")
  if term_env_var and (term_env_var.startswith("screen") or term_env_var.startswith("tmux")):
    return "\033Ptmux;\033\033]"
  return "\033]"


def _get_end_sequence():
  term_env_var = os.getenv("TERM")
  if term_env_var and (term_env_var.startswith("screen") or term_env_var.startswith("tmux")):
    return "\a\033\\"
  return "\a"


def preview(filepath: str, img_height_px: int):
  with read_image(filepath) as img:
    if img_height_px >= img.height:
      width_px, height_px = img.width, img.height
    else:
      width_px, height_px = int(img_height_px * img.width / img.height), img_height_px
    img = img.resize((width_px, height_px), Image.LANCZOS)  # type: ignore
    buffer = BytesIO()
    img.convert("RGB").save(buffer, format="JPEG")
  img_bytes = buffer.getvalue()
  img_str = base64.b64encode(img_bytes).decode("utf-8")
  print(
    f"{_get_start_sequence()}1337;"
    f"File=inline=1;size={len(img_bytes)};preserveAspectRatio=1;"
    f"width={width_px}px;height={height_px}px:{img_str}{_get_end_sequence()}",
  )
